def properties_or_ranges_have_value_type(
    elements: Iterable[aas_types.SubmodelElement], value_type: aas_types.DataTypeDefXSD
) -> bool:
    """
    Check that :paramref:`elements` which are
    :py:class:`.types.Property` or :py:class:`.types.Range`
    have the given :paramref:`value_type`.
    """
    range_or_property = (aas_types.Property, aas_types.Range)
    for element in elements:
        if isinstance(element, range_or_property):
            if element.value_type is not value_type:
                return False

    return True



def lang_strings_have_unique_languages(
    lang_strings: Iterable[aas_types.AbstractLangString],
) -> bool:
    """
    Check that :paramref:`lang_strings` are specified each for a unique
    language.
    """
    language_set = set()  # type: Set[str]
    for lang_string in lang_strings:
        if lang_string.language in language_set:
            return False

        language_set.add(lang_string.language)

    return True



"""Manipulate, verify and de/serialize Asset Administration Shells v3.1."""

# Synchronize with __init__.py and changelog.rst!
__version__ = "1.0.0"
__author__ = "Marko Ristin"
__copyright__ = "2025 Contributors to aas-core3.1-python"
__license__ = "MIT"
__status__ = "Production/Stable"



def submodel_elements_have_identical_semantic_ids(
    elements: Iterable[aas_types.SubmodelElement],
) -> bool:
    """
    Check that all :paramref:`elements` have the identical
    :py:attr:`.types.HasSemantics.semantic_id`.
    """
    that_semantic_id = None  # type: Optional[aas_types.Reference]

    for element in elements:
        if element.semantic_id is None:
            continue

        if that_semantic_id is None:
            that_semantic_id = element.semantic_id
            continue

        this_semantic_id = element.semantic_id

        if len(that_semantic_id.keys) != len(this_semantic_id.keys):
            return False

        for this_key, that_key in zip(this_semantic_id.keys, that_semantic_id.keys):
            if this_key.value != that_key.value:
                return False

    return True



def _is_leap_year(year: int) -> bool:
    """
    Check if :paramref:`year` is a leap year.

    >>> _is_leap_year(2016)
    True

    >>> _is_leap_year(1700)
    False

    >>> _is_leap_year(1600)
    True

    >>> _is_leap_year(2000)
    True
    """
    # We consider the years B.C. to be one-off.
    #
    # See the note at: https://www.w3.org/TR/xmlschema-2/#dateTime:
    # "'-0001' is the lexical representation of the year 1 Before Common Era
    # (1 BCE, sometimes written "1 BC")."
    #
    # Hence, -1 year in XML is 1 BCE, which is 0 year in astronomical years.
    if year < 0:
        year = abs(year) - 1

    # See: https://en.wikipedia.org/wiki/Leap_year#Algorithm
    if year % 4 > 0:
        return False

    if year % 100 > 0:
        return True

    if year % 400 > 0:
        return False

    return True


_DAYS_IN_MONTH: Mapping[int, int] = {
    1: 31,
    # Please use _is_leap_year if you need to check
    # whether a concrete February has 28 or 29 days.
    2: 29,
    3: 31,
    4: 30,
    5: 31,
    6: 30,
    7: 31,
    8: 31,
    9: 30,
    10: 31,
    11: 30,
    12: 31,
}


_DATE_PREFIX_RE = re.compile(r"^(-?[0-9]+)-([0-9]{2})-([0-9]{2})")


def is_xs_date(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:date``."""
    if not matches_xs_date(value):
        return False

    # NOTE (mristin, 2022-11-23):
    # We can not use :py:func:`datetime.datetime.strptime` as it does not
    # handle years below 1000 correctly on Windows (*e.g.*, ``-999-01-01``).

    # NOTE (mristin, 2022-10-30):
    # We need to match the prefix as zone offsets are allowed in the dates. Optimally,
    # we would re-use the pattern matching from :py:func`matches_xs_date`, but this
    # would make the code generation and constraint inference for schemas much more
    # difficult. Hence, we sacrifice the efficiency a bit for the clearer code & code
    # generation.
    match = _DATE_PREFIX_RE.match(value)
    assert match is not None

    year = int(match.group(1))
    month = int(match.group(2))
    day = int(match.group(3))

    # We do not accept year zero,
    # see the note at: https://www.w3.org/TR/xmlschema-2/#dateTime
    if year == 0:
        return False

    if day <= 0:
        return False

    if month <= 0 or month >= 13:
        return False

    if month == 2:
        max_days = 29 if _is_leap_year(year) else 28
    else:
        max_days = _DAYS_IN_MONTH[month]

    if day > max_days:
        return False

    return True


def is_xs_double(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:double``."""
    # We need to check explicitly for the regular expression since
    # ``float(.)`` is too permissive. For example,
    # it accepts "nan" although only "NaN" is valid.
    # See: https://www.w3.org/TR/xmlschema-2/#double
    if not matches_xs_double(value):
        return False

    converted = float(value)

    # Check that the value is either "INF" or "-INF".
    # Otherwise, the value is a decimal which is too big
    # to be represented as a double-precision floating point
    # number.
    #
    # Python simply rounds up/down to ``INF`` and ``-INF``,
    # respectively, if the number is too large.
    # For example: ``float("1e400") == math.inf``
    if math.isinf(converted) and value != "INF" and value != "-INF":
        return False

    return True


def is_xs_float(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:float``."""
    # We need to check explicitly for the regular expression since
    # ``float(.)`` is too permissive. For example,
    # it accepts "nan" although only "NaN" is valid.
    # See: https://www.w3.org/TR/xmlschema-2/#double
    if not matches_xs_float(value):
        return False

    converted = float(value)

    # Check that the value is either "INF" or "-INF".
    # Otherwise, the value is a decimal which is too big
    # to be represented as a single-precision floating point
    # number.
    #
    # Python simply rounds up/down to ``INF`` and ``-INF``,
    # respectively, if the number is too large.
    # For example: ``float("1e400") == math.inf``
    if math.isinf(converted) and value != "INF" and value != "-INF":
        return False

    # Python uses double-precision floating point numbers. Since
    # we check for a single-precision one, we have to explicitly
    # see if the number is within a range of a single-precision
    # floating point numbers.
    try:
        _ = struct.pack(">f", converted)
    except OverflowError:
        return False

    return True


def is_xs_g_month_day(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:gMonthDay``."""
    if not matches_xs_g_month_day(value):
        return False

    month = int(value[2:4])
    day = int(value[5:7])

    max_days = _DAYS_IN_MONTH[month]
    return day <= max_days


def is_xs_long(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:long``."""
    if not matches_xs_long(value):
        return False

    converted = int(value)
    return -9223372036854775808 <= converted <= 9223372036854775807


def is_xs_int(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:int``."""
    if not matches_xs_int(value):
        return False

    converted = int(value)
    return -2147483648 <= converted <= 2147483647


def is_xs_short(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:short``."""
    if not matches_xs_short(value):
        return False

    converted = int(value)
    return -32768 <= converted <= 32767


def is_xs_byte(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:byte``."""
    if not matches_xs_byte(value):
        return False

    converted = int(value)
    return -128 <= converted <= 127


def is_xs_unsigned_long(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:unsignedLong``."""
    if not matches_xs_unsigned_long(value):
        return False

    converted = int(value)
    return 0 <= converted <= 18446744073709551615


def is_xs_unsigned_int(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:unsignedInt``."""
    if not matches_xs_unsigned_int(value):
        return False

    converted = int(value)
    return 0 <= converted <= 4294967295


def is_xs_unsigned_short(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:unsignedShort``."""
    if not matches_xs_unsigned_short(value):
        return False

    converted = int(value)
    return 0 <= converted <= 65535


def is_xs_unsigned_byte(value: str) -> bool:
    """Check that :paramref:`value` is a valid ``xs:unsignedByte``."""
    if not matches_xs_unsigned_byte(value):
        return False

    converted = int(value)
    return 0 <= converted <= 255


_DATA_TYPE_DEF_XSD_TO_VALUE_CONSISTENCY: Mapping[
    aas_types.DataTypeDefXSD, Callable[[str], bool]
] = {
    aas_types.DataTypeDefXSD.ANY_URI: matches_xs_any_uri,
    aas_types.DataTypeDefXSD.BASE_64_BINARY: matches_xs_base_64_binary,
    aas_types.DataTypeDefXSD.BOOLEAN: matches_xs_boolean,
    aas_types.DataTypeDefXSD.BYTE: is_xs_byte,
    aas_types.DataTypeDefXSD.DATE: is_xs_date,
    aas_types.DataTypeDefXSD.DATE_TIME: is_xs_date_time,
    aas_types.DataTypeDefXSD.DECIMAL: matches_xs_decimal,
    aas_types.DataTypeDefXSD.DOUBLE: is_xs_double,
    aas_types.DataTypeDefXSD.DURATION: matches_xs_duration,
    aas_types.DataTypeDefXSD.FLOAT: is_xs_float,
    aas_types.DataTypeDefXSD.G_DAY: matches_xs_g_day,
    aas_types.DataTypeDefXSD.G_MONTH: matches_xs_g_month,
    aas_types.DataTypeDefXSD.G_MONTH_DAY: is_xs_g_month_day,
    aas_types.DataTypeDefXSD.G_YEAR: matches_xs_g_year,
    aas_types.DataTypeDefXSD.G_YEAR_MONTH: matches_xs_g_year_month,
    aas_types.DataTypeDefXSD.HEX_BINARY: matches_xs_hex_binary,
    aas_types.DataTypeDefXSD.INT: is_xs_int,
    aas_types.DataTypeDefXSD.INTEGER: matches_xs_integer,
    aas_types.DataTypeDefXSD.LONG: is_xs_long,
    aas_types.DataTypeDefXSD.NEGATIVE_INTEGER: matches_xs_negative_integer,
    aas_types.DataTypeDefXSD.NON_NEGATIVE_INTEGER: matches_xs_non_negative_integer,
    aas_types.DataTypeDefXSD.NON_POSITIVE_INTEGER: matches_xs_non_positive_integer,
    aas_types.DataTypeDefXSD.POSITIVE_INTEGER: matches_xs_positive_integer,
    aas_types.DataTypeDefXSD.SHORT: is_xs_short,
    aas_types.DataTypeDefXSD.STRING: matches_xs_string,
    aas_types.DataTypeDefXSD.TIME: matches_xs_time,
    aas_types.DataTypeDefXSD.UNSIGNED_BYTE: is_xs_unsigned_byte,
    aas_types.DataTypeDefXSD.UNSIGNED_INT: is_xs_unsigned_int,
    aas_types.DataTypeDefXSD.UNSIGNED_LONG: is_xs_unsigned_long,
    aas_types.DataTypeDefXSD.UNSIGNED_SHORT: is_xs_unsigned_short,
}
assert all(
    data_type_def_xsd in _DATA_TYPE_DEF_XSD_TO_VALUE_CONSISTENCY
    for data_type_def_xsd in aas_types.DataTypeDefXSD
)


def value_consistent_with_xsd_type(
    value: str, value_type: aas_types.DataTypeDefXSD
) -> bool:
    """
    Check that :paramref:`value` is consistent with the given
    :paramref:`value_type`.
    """
    return _DATA_TYPE_DEF_XSD_TO_VALUE_CONSISTENCY[value_type](value)



"""Download the meta-model to this directory."""

import pathlib
import os
import argparse
import sys
from typing import Dict, Union

import requests


GITHUB_API = "https://api.github.com"
OWNER = "aas-core-works"
REPO = "aas-core-meta"
REF = "main"
FILEPATH = "aas_core_meta/v3_1.py"


def _latest_commit_sha_for_path(
    owner: str, repo: str, ref: str, path: str, *, timeout: float = 15.0
) -> str:
    """
    Resolve the latest commit SHA on a ref for a specific path.

    Uses: GET /repos/{owner}/{repo}/commits?path=...&sha=...&per_page=1
    """
    url = f"{GITHUB_API}/repos/{owner}/{repo}/commits"
    params: Dict[str, Union[str, int]] = {"path": path, "sha": ref, "per_page": 1}
    resp = requests.get(url, params=params, timeout=timeout)
    try:
        resp.raise_for_status()
    except requests.HTTPError as ex:
        raise RuntimeError(
            f"Failed to resolve latest commit for {owner}/{repo}:{ref} path={path} "
            f"({resp.status_code}): {resp.text}"
        ) from ex

    commits = resp.json()
    if not commits or not isinstance(commits, list):
        raise RuntimeError(
            "Could not determine latest commit SHA (empty API response)."
        )

    sha = commits[0].get("sha")
    if not sha or not isinstance(sha, str):
        raise RuntimeError("API did not return a valid commit SHA.")
    assert isinstance(sha, str)
    return sha


def _download_raw_from_commit(
    owner: str, repo: str, sha: str, path: str, *, timeout: float = 30.0
) -> str:
    """
    Download the raw file contents for a given commit SHA and path.
    """
    raw_url = f"https://raw.githubusercontent.com/{owner}/{repo}/{sha}/{path}"
    resp = requests.get(raw_url, timeout=timeout)
    try:
        resp.raise_for_status()
    except requests.HTTPError as ex:
        raise RuntimeError(
            f"Failed to download file from {raw_url} ({resp.status_code}): {resp.text}"
        ) from ex
    # We expect Python source, decode as text.
    return resp.text


def main() -> int:
    """Execute the main routine."""
    parser = argparse.ArgumentParser(description=__doc__)
    _ = parser.parse_args()

    this_dir = pathlib.Path(os.path.realpath(__file__)).parent

    # 1) Get the latest commit SHA for the file on `main`.
    sha = _latest_commit_sha_for_path(OWNER, REPO, REF, FILEPATH)

    # 2) Download the exact file from that commit.
    content = _download_raw_from_commit(OWNER, REPO, sha, FILEPATH)

    # 3) Write to current_dir/v3_1.py, with the required comment
    #    at the beginning *and* at the end.
    pinned_url = f"https://raw.githubusercontent.com/{OWNER}/{REPO}/{sha}/{FILEPATH}"
    banner = f"# Downloaded from: {pinned_url}\n"

    out_path = this_dir / "meta_model.py"
    # Overwrite unconditionally.
    out_path.write_text(f"{banner}{content.rstrip()}\n\n{banner}", encoding="utf-8")

    print(f"Wrote {out_path} (from commit {sha[:12]}).")
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Provide functionality used across the tests."""

import base64
import collections.abc
import difflib
import enum
import io
import os
import pathlib
import textwrap
from typing import Union, Sequence

import aas_core3_1.common as aas_common
import aas_core3_1.types as aas_types

_REPO_ROOT = pathlib.Path(os.path.realpath(__file__)).parent.parent

#: Path to the directory which contains input and golden files
TEST_DATA_DIR = _REPO_ROOT / "test_data"

#: If set, the golden files in the tests should be re-recorded instead
#: of checked against.
RECORD_MODE = os.environ.get("AAS_CORE3_1_PYTHON_TESTS_RECORD_MODE", "").lower() in (
    "1",
    "on",
    "true",
)


def record_or_check(path: pathlib.Path, got: str) -> None:
    """
    Re-record or check that :paramref:`got` content matches the content of
    :paramref:`path`.

    If :py:attr:`~RECORD_MODE` is set, the content of :paramref:`path` will be
    simply overwritten with :paramref:`got` content, and no checks are performed.

    :param path: to the golden file
    :param got: obtained content
    :raise: :py:class:`AssertionError` if the contents do not match
    """
    if RECORD_MODE:
        path.parent.mkdir(exist_ok=True, parents=True)
        path.write_text(got, encoding="utf-8")
    else:
        if not path.exists():
            raise FileNotFoundError(
                f"The golden file could not be found: {path}; did you record it?"
            )

        expected = path.read_text(encoding="utf-8")
        if expected != got:
            writer = io.StringIO()

            diff = difflib.ndiff(
                expected.splitlines(keepends=True), got.splitlines(keepends=True)
            )

            diff_text = "".join(diff)
            writer.write(
                f"""\
The obtained content and the content of {path} do not match:
{diff_text}"""
            )

            raise AssertionError(writer.getvalue())


def trace(
    that: Union[
        bool,
        int,
        float,
        str,
        bytes,
        enum.Enum,
        aas_types.Class,
        Sequence[aas_types.Class],
    ],
) -> str:
    """
    Generate a segment in a trace of an iteration.

    :param that: to be traced
    :return: segment in the descent trace
    """
    if isinstance(that, aas_types.Class):
        if isinstance(that, aas_types.Identifiable):
            return f"{that.__class__.__name__} with ID {that.id}"
        elif isinstance(that, aas_types.Referable):
            return f"{that.__class__.__name__} with ID-short {that.id_short}"
        else:
            return that.__class__.__name__
    elif isinstance(that, (bool, int, float, str, enum.Enum)):
        return str(that)
    elif isinstance(that, bytes):
        return base64.b64encode(that).decode("ascii")
    elif isinstance(that, collections.abc.Sequence):
        if len(that) == 0:
            return "[]"
        writer = io.StringIO()
        writer.write("[\n")
        for i, item in enumerate(that):
            assert isinstance(item, aas_types.Class)

            writer.write(textwrap.indent(trace(item), "  "))

            if i < len(that) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write("]")
        return writer.getvalue()
    else:
        aas_common.assert_never(that)


def trace_log_as_text_file_content(log: Sequence[str]) -> str:
    """
    Convert the trace log to a text to be stored in a file.

    :param log: to be converted to text
    :return: content of the file, including the new-line at the end
    """
    writer = io.StringIO()
    for entry in log:
        writer.write(f"{entry}\n")
    writer.write("\n")
    return writer.getvalue()



def id_shorts_are_unique(referables: Iterable[aas_types.Referable]) -> bool:
    """
    Check that all :py:attr:`.types.Referable.id_short` are unique
    among :paramref:`referables`.
    """
    id_short_set = set()  # type: Set[str]
    for referable in referables:
        if referable.id_short in id_short_set:
            return False

        if referable.id_short is not None:
            id_short_set.add(referable.id_short)

    return True



def extension_names_are_unique(extensions: Iterable[aas_types.Extension]) -> bool:
    """
    Check that all :py:attr:`.types.Extension.name` are unique
    among :paramref:`extensions`.
    """
    name_set = set()  # type: Set[str]
    for extension in extensions:
        if extension.name in name_set:
            return False

        name_set.add(extension.name)

    return True



def value_type_or_default(self) -> "DataTypeDefXSD":
    """Return the :py:attr:`value_type` if set, or the default otherwise."""
    return self.value_type if self.value_type is not None else DataTypeDefXSD.STRING



"""
Update everything in this project to the latest aas-core-meta and -codegen.

Git is expected to be installed.
"""

from __future__ import annotations

import argparse
import os
import pathlib
import re
import subprocess
import sys
import tempfile
import time
from typing import Optional, List, Callable, AnyStr, Sequence


def _make_sure_no_changed_files(
    repo_dir: pathlib.Path, expected_branch: str
) -> Optional[int]:
    """
    Make sure that no files are modified in the given repository.

    Return exit code if something is unexpected.
    """
    diff_name_status = subprocess.check_output(
        ["git", "diff", "--name-status", expected_branch],
        cwd=str(repo_dir),
        encoding="utf-8",
    ).strip()

    if len(diff_name_status.splitlines()) > 0:
        print(
            f"The following files are modified "
            f"compared to branch {expected_branch!r} in {repo_dir}:\n"
            f"{diff_name_status}\n"
            f"\n"
            f"Please stash the changes first before you update to aas-core-meta.",
            file=sys.stderr,
        )
        return 1

    return None


def _run_in_parallel(
    calls: Sequence[Callable[[], subprocess.Popen[AnyStr]]],
    on_status_update: Callable[[int], None],
) -> Optional[int]:
    """
    Run the given scripts in parallel.

    Return an error code, if any.
    """
    procs = []  # type: List[subprocess.Popen[AnyStr]]

    try:
        for call in calls:
            proc = call()
            procs.append(proc)

        failure = False
        remaining_procs = sum(1 for proc in procs if proc.returncode is None)

        next_print = time.time() + 15
        while remaining_procs > 0:
            if time.time() > next_print:
                on_status_update(remaining_procs)
                next_print = time.time() + 15

            time.sleep(1)

            for proc in procs:
                proc.poll()

                if proc.returncode is not None:
                    if proc.returncode != 0:
                        failure = True

            if failure:
                print(
                    "One or more processes failed. Terminating all the processes...",
                    file=sys.stderr,
                )
                for proc in procs:
                    proc.terminate()

                print("Terminated all the processes.", file=sys.stderr)
                return 1

            for proc in procs:
                proc.poll()

            remaining_procs = sum(1 for proc in procs if proc.returncode is None)

        return None
    finally:
        for proc in procs:
            if proc.returncode is None:
                proc.terminate()


def _regenerate_code(our_repo: pathlib.Path) -> Optional[int]:
    """
    Call codegen script.

    Return an error code, if any.
    """
    codegen_dir = our_repo / "dev_scripts/codegen"

    meta_model_path = codegen_dir / "meta_model.py"

    target_dir = our_repo

    print("Starting to run codegen script")
    start = time.perf_counter()

    proc = subprocess.run(
        [
            sys.executable,
            "codegen.py",
            "--meta_model",
            str(meta_model_path),
            "--target",
            str(target_dir),
        ],
        cwd=str(codegen_dir),
        check=True,
    )

    if proc.returncode != 0:
        return proc.returncode

    duration = time.perf_counter() - start
    print(f"Generating the code took: {duration:.2f} seconds.")

    return None


def _reformat_code(our_repo: pathlib.Path) -> None:
    """Reformat the generated code."""
    print("Re-formatting the code...")

    precommit_script = our_repo / "continuous_integration/precommit.py"

    subprocess.check_call(
        [sys.executable, str(precommit_script), "--select", "reformat", "--overwrite"],
        cwd=our_repo,
    )


def _run_tests_and_rerecord(our_repo: pathlib.Path) -> Optional[int]:
    """
    Run the tests with the environment variables set to re-record.

    Return the error code, if any.
    """
    print("Running tests & re-recording the test traces...")

    env = os.environ.copy()
    env["AAS_CORE3_1_PYTHON_TESTS_RECORD_MODE"] = "true"

    # NOTE (mristin):
    # We need to include the repository root on the PYTHNPATH since the newer
    # versions of Python (such as 3.11 and 3.12) exclude ``tests/`` from it --
    # they rely on pyproject.toml excluding them in ``find_package``:
    #
    # ``packages=find_packages(exclude=["tests", ...]),``
    #
    # . This means that the ``tests`` module will not be on the Python path, as newer
    # versions of setuptools only put packages explicitly found by ``find_packages``.

    python_path = env.get("PYTHONPATH", None)
    if python_path is None:
        python_path = str(our_repo)
    else:
        python_path = f"{python_path}{os.pathsep}{str(our_repo)}"

    env["PYTHONPATH"] = python_path

    test_files = sorted((our_repo / "tests").glob("**/test_*.py"))

    # pylint: disable=consider-using-with
    calls = [
        lambda a_pth=pth, cwd=our_repo: subprocess.Popen(  # type: ignore
            [sys.executable, str(a_pth)],
            cwd=str(cwd),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            encoding="utf-8",
            env=env,
        )
        for pth in test_files
    ]  # type: Sequence[Callable[[], subprocess.Popen[str]]]
    # pylint: enable=consider-using-with

    test_files_joined = ",\n".join(str(pth) for pth in test_files)
    print(f"Starting to run test modules:\n{test_files_joined}")
    start = time.perf_counter()

    exit_code = _run_in_parallel(
        calls=calls,
        on_status_update=(
            lambda remaining: print(
                f"There are {remaining} test module(s) still running..."
            )
        ),
    )
    if exit_code is not None:
        return exit_code

    duration = time.perf_counter() - start
    print(f"Re-recording took: {duration:.2f} seconds.")

    return None


def _create_branch_commit_and_push(
    our_repo: pathlib.Path,
    aas_core_meta_revision: str,
    aas_core_codegen_revision: str,
) -> None:
    """Create a feature branch, commit the changes and push it."""
    branch = (
        f"Update-to-aas-core-meta-codegen-{aas_core_meta_revision}-"
        f"{aas_core_codegen_revision}"
    )
    print(f"Creating the branch {branch!r}...")
    subprocess.check_call(["git", "checkout", "-b", branch], cwd=our_repo)

    print("Adding files...")
    subprocess.check_call(["git", "add", "."], cwd=our_repo)

    # pylint: disable=line-too-long
    message = f"""\
Update to aas-core-meta, codegen {aas_core_meta_revision}, {aas_core_codegen_revision}

We update the development requirements to and re-generate everything
with:
* [aas-core-meta {aas_core_meta_revision}] and
* [aas-core-codegen {aas_core_codegen_revision}].

[aas-core-meta {aas_core_meta_revision}]: https://github.com/aas-core-works/aas-core-meta/commit/{aas_core_meta_revision}
[aas-core-codegen {aas_core_codegen_revision}]: https://github.com/aas-core-works/aas-core-codegen/commit/{aas_core_codegen_revision}
"""

    # pylint: enable=line-too-long

    print("Committing...")
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_file = pathlib.Path(tmp_dir) / "commit-message.txt"
        tmp_file.write_text(message, encoding="utf-8")

        subprocess.check_call(["git", "commit", "--file", str(tmp_file)], cwd=our_repo)

    print(f"Pushing to remote {branch}...")
    subprocess.check_call(["git", "push", "-u"], cwd=our_repo)


_AAS_CORE_CODEGEN_SHA_RE = re.compile(
    r"aas-core-codegen@git\+https://github.com/aas-core-works/aas-core-codegen@([a-zA-Z0-9]+)"
)


def _get_codegen_revision(our_repo: pathlib.Path) -> str | None:
    pyproject_toml_path = our_repo / "dev_scripts/pyproject.toml"

    codegen_sha: str | None = None

    sha_re = re.compile(_AAS_CORE_CODEGEN_SHA_RE)

    try:
        with pyproject_toml_path.open("r") as pyproject_toml_file:
            for line in pyproject_toml_file:
                matches = sha_re.search(line)

                if matches is None:
                    continue

                codegen_sha = matches.group(1)
                break

    except OSError as os_error:
        print(f"Cannot read codegen revision: {os_error}.")

    if codegen_sha is None:
        print("Cannot read codegen revision.")

    return codegen_sha


_AAS_CORE_META_SHA_RE = re.compile(
    r"https://raw.githubusercontent.com/aas-core-works/aas-core-meta/([a-zA-Z0-9]+)/aas_core_meta/v.*.py"
)


def _get_meta_model_revision(our_repo: pathlib.Path) -> str | None:
    meta_model_path = our_repo / "dev_scripts/codegen/meta_model.py"

    meta_model_sha: str | None = None

    sha_re = re.compile(_AAS_CORE_META_SHA_RE)

    try:
        with meta_model_path.open("r") as meta_model_file:
            for line in meta_model_file:
                matches = sha_re.search(line)

                if matches is None:
                    continue

                meta_model_sha = matches.group(1)[:8]
                break

    except OSError as os_error:
        print(f"Cannot read meta model revision: {os_error}.")

    if meta_model_sha is None:
        print("Cannot read meta model revision.")

    return meta_model_sha


def main() -> int:
    """Execute the main routine."""
    this_path = pathlib.Path(os.path.realpath(__file__))
    our_repo = this_path.parent.parent

    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        "--expected_our_branch",
        help="Git branch expected in this repository",
        default="main",
    )

    args = parser.parse_args()

    expected_our_branch = str(args.expected_our_branch)

    # region Our repo

    our_branch = subprocess.check_output(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"],
        cwd=str(our_repo),
        encoding="utf-8",
    ).strip()
    if our_branch != expected_our_branch:
        print(
            f"--expected_our_branch is {expected_our_branch}, "
            f"but got {our_branch} in: {our_repo}",
            file=sys.stderr,
        )
        return 1

    # endregion

    exit_code = _make_sure_no_changed_files(
        repo_dir=our_repo, expected_branch=expected_our_branch
    )
    if exit_code is not None:
        return exit_code

    exit_code = _regenerate_code(our_repo=our_repo)
    if exit_code is not None:
        return exit_code

    _reformat_code(our_repo=our_repo)

    exit_code = _run_tests_and_rerecord(our_repo=our_repo)
    if exit_code is not None:
        return exit_code

    aas_core_codegen_revision = _get_codegen_revision(our_repo=our_repo)
    if aas_core_codegen_revision is None:
        return 1

    aas_core_meta_revision = _get_meta_model_revision(our_repo=our_repo)
    if aas_core_meta_revision is None:
        return 1

    _create_branch_commit_and_push(
        our_repo=our_repo,
        aas_core_meta_revision=aas_core_meta_revision,
        aas_core_codegen_revision=aas_core_codegen_revision,
    )

    return 0


if __name__ == "__main__":
    sys.exit(main())



def is_xs_date_time(value: str) -> bool:
    """
    Check that :paramref:`value` is a ``xs:dateTime`` with
    the time zone set to UTC.
    """
    if not matches_xs_date_time(value):
        return False

    date, _ = value.split("T")
    return is_xs_date(date)



def qualifier_types_are_unique(qualifiers: Iterable[aas_types.Qualifier]) -> bool:
    """
    Check that there are no duplicate
    :py:attr:`.types.Qualifier.type`'s
    in the :paramref:`qualifiers`.
    """
    type_set = set()  # type: Set[str]
    for qualifier in qualifiers:
        if qualifier.type in type_set:
            return False

        type_set.add(qualifier.type)

    return True



"""Generate the SDK based on the aas-core-meta model and the snippets."""

import argparse
import os
import pathlib
import subprocess
import sys


def _generate_sdk(
    meta_model_path: pathlib.Path,
    snippet_path: pathlib.Path,
    sdk_path: pathlib.Path,
) -> None:
    subprocess.run(
        [
            "aas-core-codegen",
            "--model_path",
            str(meta_model_path),
            "--snippets_dir",
            str(snippet_path),
            "--output_dir",
            str(sdk_path / "aas_core3_1"),
            "--target",
            "python",
        ],
        check=True,
    )


def main() -> int:
    """Execute the main routine."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--meta_model", type=pathlib.Path, required=True)
    parser.add_argument("--target", type=pathlib.Path, required=True)
    args = parser.parse_args()

    this_dir = pathlib.Path(os.path.realpath(__file__)).parent

    snippet_dir = this_dir / "snippets"

    _generate_sdk(args.meta_model, snippet_dir, this_dir.parent.parent)

    return 0


if __name__ == "__main__":
    sys.exit(main())



# This code has been automatically generated by:
# dev_scripts/test_codegen/generate_common_jsonization.py
# Do NOT edit or append.


"""Provide common functionality for JSON de-serialization."""


# pylint: disable=missing-docstring


import collections.abc
import json
import sys
from typing import Iterator, Optional

if sys.version_info >= (3, 8):
    from typing import Final
else:
    from typing_extensions import Final

import aas_core3_1.common as aas_common
import aas_core3_1.jsonization as aas_jsonization
import aas_core3_1.types as aas_types

import tests.common


class Difference:
    """Represent a single difference between two JSON-ables."""

    #: Human-readable description of the difference
    message: Final[str]

    #: Path in the expected JSON-able value which is different from
    #: the obtained JSON-able value
    path: Final[aas_jsonization.Path]

    def __init__(self, message: str) -> None:
        """Initialize with the given message and empty path."""
        self.message = message
        self.path = aas_jsonization.Path()

    def __str__(self) -> str:
        return f"#{self.path}: {self.message}"


def check_equal(
    expected: aas_jsonization.Jsonable,
    got: aas_jsonization.Jsonable,
) -> Iterator[Difference]:
    """
    Compare recursively two JSON-able values for equality.

    :param expected: expected JSON-able value
    :param got: obtained JSON-able value
    :yield: differences
    """
    if isinstance(expected, (bool, int, float, str, bytes)):
        if type(expected) != type(got):  # pylint: disable=unidiomatic-typecheck
            yield Difference(f"Expected {type(expected)}, but got {type(got)}")

        if expected != got:
            yield Difference(f"Expected {expected!r}, but got {got!r}")
    elif isinstance(expected, collections.abc.Sequence):
        if not isinstance(got, collections.abc.Sequence):
            yield Difference(f"Expected a sequence, but got {type(got)}")
        else:
            if len(expected) != len(got):
                yield Difference(
                    f"Expected a sequence of {len(expected)} item(s), "
                    f"but got {len(got)} item(s)"
                )

            for i, (expected_item, got_item) in enumerate(zip(expected, got)):
                for difference in check_equal(expected_item, got_item):
                    difference.path._prepend(aas_jsonization.IndexSegment(expected, i))
                    yield difference

    elif isinstance(expected, collections.abc.Mapping):
        if not isinstance(got, collections.abc.Mapping):
            yield Difference(f"Expected a mapping, but got {type(got)}")
        else:
            if not all(isinstance(key, str) for key in expected.keys()):
                raise ValueError(
                    f"Expected all keys in the expected JSON-able value to be strings, "
                    f"but got: {list(expected.keys())}"
                )

            if not all(isinstance(key, str) for key in got.keys()):
                yield Difference(
                    f"Expected all keys in a mapping to be strings, "
                    f"but got: {list(got.keys())}"
                )

            expected_key_set = set(expected.keys())
            got_key_set = set(got.keys())

            expected_got_diff = expected_key_set.difference(got_key_set)
            if expected_got_diff:
                yield Difference(f"Expected key(s) {sorted(expected_got_diff)} missing")

            got_expected_diff = got_key_set.difference(expected_key_set)
            if got_expected_diff:
                yield Difference(f"Unexpected key(s) {sorted(got_expected_diff)}")

            for key, expected_value in expected.items():
                got_value = got[key]

                for difference in check_equal(expected_value, got_value):
                    difference.path._prepend(
                        aas_jsonization.PropertySegment(expected, key)
                    )
                    yield difference
    else:
        aas_common.assert_never(expected)


def load_maximal_extension() -> aas_types.Extension:
    """
    Load a maximal example
    of :py:class:`aas_types.Extension`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Extension"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Extension] = None

    for something in container.descend():
        if isinstance(something, aas_types.Extension):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Extension " f"in {path}, but found none."
        )

    return instance


def load_minimal_extension() -> aas_types.Extension:
    """
    Load a minimal example
    of :py:class:`aas_types.Extension`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Extension"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Extension] = None

    for something in container.descend():
        if isinstance(something, aas_types.Extension):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Extension " f"in {path}, but found none."
        )

    return instance


def load_maximal_administrative_information() -> aas_types.AdministrativeInformation:
    """
    Load a maximal example
    of :py:class:`aas_types.AdministrativeInformation`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "AdministrativeInformation"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.AdministrativeInformation] = None

    for something in container.descend():
        if isinstance(something, aas_types.AdministrativeInformation):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of AdministrativeInformation "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_administrative_information() -> aas_types.AdministrativeInformation:
    """
    Load a minimal example
    of :py:class:`aas_types.AdministrativeInformation`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "AdministrativeInformation"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.AdministrativeInformation] = None

    for something in container.descend():
        if isinstance(something, aas_types.AdministrativeInformation):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of AdministrativeInformation "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_qualifier() -> aas_types.Qualifier:
    """
    Load a maximal example
    of :py:class:`aas_types.Qualifier`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Qualifier"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Qualifier] = None

    for something in container.descend():
        if isinstance(something, aas_types.Qualifier):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Qualifier " f"in {path}, but found none."
        )

    return instance


def load_minimal_qualifier() -> aas_types.Qualifier:
    """
    Load a minimal example
    of :py:class:`aas_types.Qualifier`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Qualifier"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Qualifier] = None

    for something in container.descend():
        if isinstance(something, aas_types.Qualifier):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Qualifier " f"in {path}, but found none."
        )

    return instance


def load_maximal_asset_administration_shell() -> aas_types.AssetAdministrationShell:
    """
    Load a maximal example
    of :py:class:`aas_types.AssetAdministrationShell`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "AssetAdministrationShell"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.AssetAdministrationShell] = None

    for something in container.descend():
        if isinstance(something, aas_types.AssetAdministrationShell):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of AssetAdministrationShell "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_asset_administration_shell() -> aas_types.AssetAdministrationShell:
    """
    Load a minimal example
    of :py:class:`aas_types.AssetAdministrationShell`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "AssetAdministrationShell"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.AssetAdministrationShell] = None

    for something in container.descend():
        if isinstance(something, aas_types.AssetAdministrationShell):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of AssetAdministrationShell "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_asset_information() -> aas_types.AssetInformation:
    """
    Load a maximal example
    of :py:class:`aas_types.AssetInformation`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "AssetInformation"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.AssetInformation] = None

    for something in container.descend():
        if isinstance(something, aas_types.AssetInformation):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of AssetInformation "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_asset_information() -> aas_types.AssetInformation:
    """
    Load a minimal example
    of :py:class:`aas_types.AssetInformation`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "AssetInformation"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.AssetInformation] = None

    for something in container.descend():
        if isinstance(something, aas_types.AssetInformation):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of AssetInformation "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_resource() -> aas_types.Resource:
    """
    Load a maximal example
    of :py:class:`aas_types.Resource`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Resource"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Resource] = None

    for something in container.descend():
        if isinstance(something, aas_types.Resource):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Resource " f"in {path}, but found none."
        )

    return instance


def load_minimal_resource() -> aas_types.Resource:
    """
    Load a minimal example
    of :py:class:`aas_types.Resource`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Resource"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Resource] = None

    for something in container.descend():
        if isinstance(something, aas_types.Resource):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Resource " f"in {path}, but found none."
        )

    return instance


def load_maximal_specific_asset_id() -> aas_types.SpecificAssetID:
    """
    Load a maximal example
    of :py:class:`aas_types.SpecificAssetID`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "SpecificAssetId"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.SpecificAssetID] = None

    for something in container.descend():
        if isinstance(something, aas_types.SpecificAssetID):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of SpecificAssetID "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_specific_asset_id() -> aas_types.SpecificAssetID:
    """
    Load a minimal example
    of :py:class:`aas_types.SpecificAssetID`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "SpecificAssetId"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.SpecificAssetID] = None

    for something in container.descend():
        if isinstance(something, aas_types.SpecificAssetID):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of SpecificAssetID "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_submodel() -> aas_types.Submodel:
    """
    Load a maximal example
    of :py:class:`aas_types.Submodel`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Submodel"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Submodel] = None

    for something in container.descend():
        if isinstance(something, aas_types.Submodel):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Submodel " f"in {path}, but found none."
        )

    return instance


def load_minimal_submodel() -> aas_types.Submodel:
    """
    Load a minimal example
    of :py:class:`aas_types.Submodel`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Submodel"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Submodel] = None

    for something in container.descend():
        if isinstance(something, aas_types.Submodel):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Submodel " f"in {path}, but found none."
        )

    return instance


def load_maximal_relationship_element() -> aas_types.RelationshipElement:
    """
    Load a maximal example
    of :py:class:`aas_types.RelationshipElement`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "RelationshipElement"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.RelationshipElement] = None

    for something in container.descend():
        if isinstance(something, aas_types.RelationshipElement):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of RelationshipElement "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_relationship_element() -> aas_types.RelationshipElement:
    """
    Load a minimal example
    of :py:class:`aas_types.RelationshipElement`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "RelationshipElement"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.RelationshipElement] = None

    for something in container.descend():
        if isinstance(something, aas_types.RelationshipElement):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of RelationshipElement "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_submodel_element_list() -> aas_types.SubmodelElementList:
    """
    Load a maximal example
    of :py:class:`aas_types.SubmodelElementList`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "SubmodelElementList"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.SubmodelElementList] = None

    for something in container.descend():
        if isinstance(something, aas_types.SubmodelElementList):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of SubmodelElementList "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_submodel_element_list() -> aas_types.SubmodelElementList:
    """
    Load a minimal example
    of :py:class:`aas_types.SubmodelElementList`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "SubmodelElementList"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.SubmodelElementList] = None

    for something in container.descend():
        if isinstance(something, aas_types.SubmodelElementList):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of SubmodelElementList "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_submodel_element_collection() -> aas_types.SubmodelElementCollection:
    """
    Load a maximal example
    of :py:class:`aas_types.SubmodelElementCollection`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "SubmodelElementCollection"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.SubmodelElementCollection] = None

    for something in container.descend():
        if isinstance(something, aas_types.SubmodelElementCollection):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of SubmodelElementCollection "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_submodel_element_collection() -> aas_types.SubmodelElementCollection:
    """
    Load a minimal example
    of :py:class:`aas_types.SubmodelElementCollection`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "SubmodelElementCollection"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.SubmodelElementCollection] = None

    for something in container.descend():
        if isinstance(something, aas_types.SubmodelElementCollection):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of SubmodelElementCollection "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_property() -> aas_types.Property:
    """
    Load a maximal example
    of :py:class:`aas_types.Property`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Property"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Property] = None

    for something in container.descend():
        if isinstance(something, aas_types.Property):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Property " f"in {path}, but found none."
        )

    return instance


def load_minimal_property() -> aas_types.Property:
    """
    Load a minimal example
    of :py:class:`aas_types.Property`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Property"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Property] = None

    for something in container.descend():
        if isinstance(something, aas_types.Property):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Property " f"in {path}, but found none."
        )

    return instance


def load_maximal_multi_language_property() -> aas_types.MultiLanguageProperty:
    """
    Load a maximal example
    of :py:class:`aas_types.MultiLanguageProperty`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "MultiLanguageProperty"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.MultiLanguageProperty] = None

    for something in container.descend():
        if isinstance(something, aas_types.MultiLanguageProperty):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of MultiLanguageProperty "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_multi_language_property() -> aas_types.MultiLanguageProperty:
    """
    Load a minimal example
    of :py:class:`aas_types.MultiLanguageProperty`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "MultiLanguageProperty"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.MultiLanguageProperty] = None

    for something in container.descend():
        if isinstance(something, aas_types.MultiLanguageProperty):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of MultiLanguageProperty "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_range() -> aas_types.Range:
    """
    Load a maximal example
    of :py:class:`aas_types.Range`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Range"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Range] = None

    for something in container.descend():
        if isinstance(something, aas_types.Range):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Range " f"in {path}, but found none."
        )

    return instance


def load_minimal_range() -> aas_types.Range:
    """
    Load a minimal example
    of :py:class:`aas_types.Range`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Range"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Range] = None

    for something in container.descend():
        if isinstance(something, aas_types.Range):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Range " f"in {path}, but found none."
        )

    return instance


def load_maximal_reference_element() -> aas_types.ReferenceElement:
    """
    Load a maximal example
    of :py:class:`aas_types.ReferenceElement`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "ReferenceElement"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.ReferenceElement] = None

    for something in container.descend():
        if isinstance(something, aas_types.ReferenceElement):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of ReferenceElement "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_reference_element() -> aas_types.ReferenceElement:
    """
    Load a minimal example
    of :py:class:`aas_types.ReferenceElement`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "ReferenceElement"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.ReferenceElement] = None

    for something in container.descend():
        if isinstance(something, aas_types.ReferenceElement):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of ReferenceElement "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_blob() -> aas_types.Blob:
    """
    Load a maximal example
    of :py:class:`aas_types.Blob`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Blob"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Blob] = None

    for something in container.descend():
        if isinstance(something, aas_types.Blob):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Blob " f"in {path}, but found none."
        )

    return instance


def load_minimal_blob() -> aas_types.Blob:
    """
    Load a minimal example
    of :py:class:`aas_types.Blob`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Blob"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Blob] = None

    for something in container.descend():
        if isinstance(something, aas_types.Blob):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Blob " f"in {path}, but found none."
        )

    return instance


def load_maximal_file() -> aas_types.File:
    """
    Load a maximal example
    of :py:class:`aas_types.File`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "File"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.File] = None

    for something in container.descend():
        if isinstance(something, aas_types.File):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of File " f"in {path}, but found none."
        )

    return instance


def load_minimal_file() -> aas_types.File:
    """
    Load a minimal example
    of :py:class:`aas_types.File`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "File"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.File] = None

    for something in container.descend():
        if isinstance(something, aas_types.File):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of File " f"in {path}, but found none."
        )

    return instance


def load_maximal_annotated_relationship_element() -> (
    aas_types.AnnotatedRelationshipElement
):
    """
    Load a maximal example
    of :py:class:`aas_types.AnnotatedRelationshipElement`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "AnnotatedRelationshipElement"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.AnnotatedRelationshipElement] = None

    for something in container.descend():
        if isinstance(something, aas_types.AnnotatedRelationshipElement):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of AnnotatedRelationshipElement "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_annotated_relationship_element() -> (
    aas_types.AnnotatedRelationshipElement
):
    """
    Load a minimal example
    of :py:class:`aas_types.AnnotatedRelationshipElement`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "AnnotatedRelationshipElement"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.AnnotatedRelationshipElement] = None

    for something in container.descend():
        if isinstance(something, aas_types.AnnotatedRelationshipElement):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of AnnotatedRelationshipElement "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_entity() -> aas_types.Entity:
    """
    Load a maximal example
    of :py:class:`aas_types.Entity`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Entity"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Entity] = None

    for something in container.descend():
        if isinstance(something, aas_types.Entity):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Entity " f"in {path}, but found none."
        )

    return instance


def load_minimal_entity() -> aas_types.Entity:
    """
    Load a minimal example
    of :py:class:`aas_types.Entity`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Entity"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Entity] = None

    for something in container.descend():
        if isinstance(something, aas_types.Entity):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Entity " f"in {path}, but found none."
        )

    return instance


def load_maximal_event_payload() -> aas_types.EventPayload:
    """
    Load a maximal example
    of :py:class:`aas_types.EventPayload`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "SelfContained"
        / "Expected"
        / "EventPayload"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    return aas_jsonization.event_payload_from_jsonable(jsonable)


def load_minimal_event_payload() -> aas_types.EventPayload:
    """
    Load a minimal example
    of :py:class:`aas_types.EventPayload`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "SelfContained"
        / "Expected"
        / "EventPayload"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    return aas_jsonization.event_payload_from_jsonable(jsonable)


def load_maximal_basic_event_element() -> aas_types.BasicEventElement:
    """
    Load a maximal example
    of :py:class:`aas_types.BasicEventElement`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "BasicEventElement"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.BasicEventElement] = None

    for something in container.descend():
        if isinstance(something, aas_types.BasicEventElement):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of BasicEventElement "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_basic_event_element() -> aas_types.BasicEventElement:
    """
    Load a minimal example
    of :py:class:`aas_types.BasicEventElement`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "BasicEventElement"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.BasicEventElement] = None

    for something in container.descend():
        if isinstance(something, aas_types.BasicEventElement):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of BasicEventElement "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_operation() -> aas_types.Operation:
    """
    Load a maximal example
    of :py:class:`aas_types.Operation`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Operation"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Operation] = None

    for something in container.descend():
        if isinstance(something, aas_types.Operation):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Operation " f"in {path}, but found none."
        )

    return instance


def load_minimal_operation() -> aas_types.Operation:
    """
    Load a minimal example
    of :py:class:`aas_types.Operation`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Operation"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Operation] = None

    for something in container.descend():
        if isinstance(something, aas_types.Operation):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Operation " f"in {path}, but found none."
        )

    return instance


def load_maximal_operation_variable() -> aas_types.OperationVariable:
    """
    Load a maximal example
    of :py:class:`aas_types.OperationVariable`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "OperationVariable"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.OperationVariable] = None

    for something in container.descend():
        if isinstance(something, aas_types.OperationVariable):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of OperationVariable "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_operation_variable() -> aas_types.OperationVariable:
    """
    Load a minimal example
    of :py:class:`aas_types.OperationVariable`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "OperationVariable"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.OperationVariable] = None

    for something in container.descend():
        if isinstance(something, aas_types.OperationVariable):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of OperationVariable "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_capability() -> aas_types.Capability:
    """
    Load a maximal example
    of :py:class:`aas_types.Capability`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Capability"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Capability] = None

    for something in container.descend():
        if isinstance(something, aas_types.Capability):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Capability " f"in {path}, but found none."
        )

    return instance


def load_minimal_capability() -> aas_types.Capability:
    """
    Load a minimal example
    of :py:class:`aas_types.Capability`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Capability"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Capability] = None

    for something in container.descend():
        if isinstance(something, aas_types.Capability):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Capability " f"in {path}, but found none."
        )

    return instance


def load_maximal_concept_description() -> aas_types.ConceptDescription:
    """
    Load a maximal example
    of :py:class:`aas_types.ConceptDescription`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "ConceptDescription"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.ConceptDescription] = None

    for something in container.descend():
        if isinstance(something, aas_types.ConceptDescription):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of ConceptDescription "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_concept_description() -> aas_types.ConceptDescription:
    """
    Load a minimal example
    of :py:class:`aas_types.ConceptDescription`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "ConceptDescription"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.ConceptDescription] = None

    for something in container.descend():
        if isinstance(something, aas_types.ConceptDescription):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of ConceptDescription "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_reference() -> aas_types.Reference:
    """
    Load a maximal example
    of :py:class:`aas_types.Reference`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Reference"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Reference] = None

    for something in container.descend():
        if isinstance(something, aas_types.Reference):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Reference " f"in {path}, but found none."
        )

    return instance


def load_minimal_reference() -> aas_types.Reference:
    """
    Load a minimal example
    of :py:class:`aas_types.Reference`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Reference"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Reference] = None

    for something in container.descend():
        if isinstance(something, aas_types.Reference):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Reference " f"in {path}, but found none."
        )

    return instance


def load_maximal_key() -> aas_types.Key:
    """
    Load a maximal example
    of :py:class:`aas_types.Key`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Key"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Key] = None

    for something in container.descend():
        if isinstance(something, aas_types.Key):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Key " f"in {path}, but found none."
        )

    return instance


def load_minimal_key() -> aas_types.Key:
    """
    Load a minimal example
    of :py:class:`aas_types.Key`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "Key"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.Key] = None

    for something in container.descend():
        if isinstance(something, aas_types.Key):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of Key " f"in {path}, but found none."
        )

    return instance


def load_maximal_lang_string_name_type() -> aas_types.LangStringNameType:
    """
    Load a maximal example
    of :py:class:`aas_types.LangStringNameType`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringNameType"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringNameType] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringNameType):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringNameType "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_lang_string_name_type() -> aas_types.LangStringNameType:
    """
    Load a minimal example
    of :py:class:`aas_types.LangStringNameType`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringNameType"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringNameType] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringNameType):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringNameType "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_lang_string_text_type() -> aas_types.LangStringTextType:
    """
    Load a maximal example
    of :py:class:`aas_types.LangStringTextType`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringTextType"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringTextType] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringTextType):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringTextType "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_lang_string_text_type() -> aas_types.LangStringTextType:
    """
    Load a minimal example
    of :py:class:`aas_types.LangStringTextType`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringTextType"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringTextType] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringTextType):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringTextType "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_environment() -> aas_types.Environment:
    """
    Load a maximal example
    of :py:class:`aas_types.Environment`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "SelfContained"
        / "Expected"
        / "Environment"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    return aas_jsonization.environment_from_jsonable(jsonable)


def load_minimal_environment() -> aas_types.Environment:
    """
    Load a minimal example
    of :py:class:`aas_types.Environment`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "SelfContained"
        / "Expected"
        / "Environment"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    return aas_jsonization.environment_from_jsonable(jsonable)


def load_maximal_embedded_data_specification() -> aas_types.EmbeddedDataSpecification:
    """
    Load a maximal example
    of :py:class:`aas_types.EmbeddedDataSpecification`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "EmbeddedDataSpecification"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.EmbeddedDataSpecification] = None

    for something in container.descend():
        if isinstance(something, aas_types.EmbeddedDataSpecification):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of EmbeddedDataSpecification "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_embedded_data_specification() -> aas_types.EmbeddedDataSpecification:
    """
    Load a minimal example
    of :py:class:`aas_types.EmbeddedDataSpecification`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "EmbeddedDataSpecification"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.EmbeddedDataSpecification] = None

    for something in container.descend():
        if isinstance(something, aas_types.EmbeddedDataSpecification):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of EmbeddedDataSpecification "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_level_type() -> aas_types.LevelType:
    """
    Load a maximal example
    of :py:class:`aas_types.LevelType`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LevelType"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LevelType] = None

    for something in container.descend():
        if isinstance(something, aas_types.LevelType):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LevelType " f"in {path}, but found none."
        )

    return instance


def load_minimal_level_type() -> aas_types.LevelType:
    """
    Load a minimal example
    of :py:class:`aas_types.LevelType`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LevelType"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LevelType] = None

    for something in container.descend():
        if isinstance(something, aas_types.LevelType):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LevelType " f"in {path}, but found none."
        )

    return instance


def load_maximal_value_reference_pair() -> aas_types.ValueReferencePair:
    """
    Load a maximal example
    of :py:class:`aas_types.ValueReferencePair`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "ValueReferencePair"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.ValueReferencePair] = None

    for something in container.descend():
        if isinstance(something, aas_types.ValueReferencePair):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of ValueReferencePair "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_value_reference_pair() -> aas_types.ValueReferencePair:
    """
    Load a minimal example
    of :py:class:`aas_types.ValueReferencePair`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "ValueReferencePair"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.ValueReferencePair] = None

    for something in container.descend():
        if isinstance(something, aas_types.ValueReferencePair):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of ValueReferencePair "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_value_list() -> aas_types.ValueList:
    """
    Load a maximal example
    of :py:class:`aas_types.ValueList`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "ValueList"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.ValueList] = None

    for something in container.descend():
        if isinstance(something, aas_types.ValueList):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of ValueList " f"in {path}, but found none."
        )

    return instance


def load_minimal_value_list() -> aas_types.ValueList:
    """
    Load a minimal example
    of :py:class:`aas_types.ValueList`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "ValueList"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.ValueList] = None

    for something in container.descend():
        if isinstance(something, aas_types.ValueList):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of ValueList " f"in {path}, but found none."
        )

    return instance


def load_maximal_lang_string_preferred_name_type_iec_61360() -> (
    aas_types.LangStringPreferredNameTypeIEC61360
):
    """
    Load a maximal example
    of :py:class:`aas_types.LangStringPreferredNameTypeIEC61360`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringPreferredNameTypeIec61360"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringPreferredNameTypeIEC61360] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringPreferredNameTypeIEC61360):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringPreferredNameTypeIEC61360 "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_lang_string_preferred_name_type_iec_61360() -> (
    aas_types.LangStringPreferredNameTypeIEC61360
):
    """
    Load a minimal example
    of :py:class:`aas_types.LangStringPreferredNameTypeIEC61360`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringPreferredNameTypeIec61360"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringPreferredNameTypeIEC61360] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringPreferredNameTypeIEC61360):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringPreferredNameTypeIEC61360 "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_lang_string_short_name_type_iec_61360() -> (
    aas_types.LangStringShortNameTypeIEC61360
):
    """
    Load a maximal example
    of :py:class:`aas_types.LangStringShortNameTypeIEC61360`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringShortNameTypeIec61360"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringShortNameTypeIEC61360] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringShortNameTypeIEC61360):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringShortNameTypeIEC61360 "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_lang_string_short_name_type_iec_61360() -> (
    aas_types.LangStringShortNameTypeIEC61360
):
    """
    Load a minimal example
    of :py:class:`aas_types.LangStringShortNameTypeIEC61360`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringShortNameTypeIec61360"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringShortNameTypeIEC61360] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringShortNameTypeIEC61360):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringShortNameTypeIEC61360 "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_lang_string_definition_type_iec_61360() -> (
    aas_types.LangStringDefinitionTypeIEC61360
):
    """
    Load a maximal example
    of :py:class:`aas_types.LangStringDefinitionTypeIEC61360`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringDefinitionTypeIec61360"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringDefinitionTypeIEC61360] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringDefinitionTypeIEC61360):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringDefinitionTypeIEC61360 "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_lang_string_definition_type_iec_61360() -> (
    aas_types.LangStringDefinitionTypeIEC61360
):
    """
    Load a minimal example
    of :py:class:`aas_types.LangStringDefinitionTypeIEC61360`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "LangStringDefinitionTypeIec61360"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.LangStringDefinitionTypeIEC61360] = None

    for something in container.descend():
        if isinstance(something, aas_types.LangStringDefinitionTypeIEC61360):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of LangStringDefinitionTypeIEC61360 "
            f"in {path}, but found none."
        )

    return instance


def load_maximal_data_specification_iec_61360() -> aas_types.DataSpecificationIEC61360:
    """
    Load a maximal example
    of :py:class:`aas_types.DataSpecificationIEC61360`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "DataSpecificationIec61360"
        / "maximal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.DataSpecificationIEC61360] = None

    for something in container.descend():
        if isinstance(something, aas_types.DataSpecificationIEC61360):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of DataSpecificationIEC61360 "
            f"in {path}, but found none."
        )

    return instance


def load_minimal_data_specification_iec_61360() -> aas_types.DataSpecificationIEC61360:
    """
    Load a minimal example
    of :py:class:`aas_types.DataSpecificationIEC61360`
    from :py:attr:`tests.common.TEST_DATA_DIR`.
    """
    path = (
        tests.common.TEST_DATA_DIR
        / "Json"
        / "ContainedInEnvironment"
        / "Expected"
        / "DataSpecificationIec61360"
        / "minimal.json"
    )

    with path.open("rt") as fid:
        try:
            jsonable = json.load(fid)
        except json.decoder.JSONDecodeError as exception:
            raise AssertionError(
                f"Unexpected non-JSON content in {path}"
            ) from exception

    container = aas_jsonization.environment_from_jsonable(jsonable)

    instance: Optional[aas_types.DataSpecificationIEC61360] = None

    for something in container.descend():
        if isinstance(something, aas_types.DataSpecificationIEC61360):
            # We pick the least deep instance, so that we can also test for
            # nested instances *etc.*
            instance = something
            break

    if instance is None:
        raise AssertionError(
            f"Expected to find an instance of DataSpecificationIEC61360 "
            f"in {path}, but found none."
        )

    return instance


# This code has been automatically generated by:
# dev_scripts/test_codegen/generate_common_jsonization.py
# Do NOT edit or append.



#!/usr/bin/env python3

"""Check that the distribution and aas_core_codegen/__init__.py are in sync."""

import os
import pathlib
import sys
from typing import Optional

import aas_core3_1

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def main() -> int:
    """Execute the main routine."""
    repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent

    pyproject_toml_pth = repo_root / "pyproject.toml"
    if not pyproject_toml_pth.exists():
        raise RuntimeError(
            f"Could not find_our_type the pyproject.toml: {pyproject_toml_pth}"
        )

    with pyproject_toml_pth.open("r") as pyproject_toml_file:
        pyproject_toml_str = pyproject_toml_file.read()

    pyproject_toml_map = tomllib.loads(pyproject_toml_str)

    success = True

    ##
    # Check basic fields
    ##

    author_names = [
        author["name"] for author in pyproject_toml_map["project"]["authors"]
    ]

    if aas_core3_1.__author__ not in author_names:
        print(
            "The author aas_core_codegen/__init__.py is not part of authors in pyproject.toml. "
            f"authors in pyproject.toml is {pyproject_toml_map['authors']}, "
            f"while the author in aas_core_codegen/__init__.py is: "
            f"{aas_core3_1.__author__}",
            file=sys.stderr,
        )
        success = False

    if pyproject_toml_map["project"]["license"] != aas_core3_1.__license__:
        print(
            f"The license in the pyproject.toml is {pyproject_toml_map['project']['license']}, "
            f"while the license in aas_core_codegen/__init__.py is: "
            f"{aas_core3_1.__license__}",
            file=sys.stderr,
        )
        success = False

    if pyproject_toml_map["project"]["description"] != aas_core3_1.__doc__:
        print(
            f"The description in the pyproject.toml is {pyproject_toml_map['project']['description']}, "
            f"while the description in aas_core_codegen/__init__.py is: "
            f"{aas_core3_1.__doc__}",
            file=sys.stderr,
        )
        success = False

    ##
    # Classifiers need special attention as there are multiple.
    ##

    # This is the map from the distribution to expected status in __init__.py.
    status_map = {
        "Development Status :: 1 - Planning": "Planning",
        "Development Status :: 2 - Pre-Alpha": "Pre-Alpha",
        "Development Status :: 3 - Alpha": "Alpha",
        "Development Status :: 4 - Beta": "Beta",
        "Development Status :: 5 - Production/Stable": "Production/Stable",
        "Development Status :: 6 - Mature": "Mature",
        "Development Status :: 7 - Inactive": "Inactive",
    }

    classifiers = pyproject_toml_map["project"]["classifiers"]

    status_classifier = None  # type: Optional[str]
    for classifier in classifiers:
        if classifier in status_map:
            status_classifier = classifier
            break

    if status_classifier is None:
        print(
            "Expected a status classifier in pyproject.toml "
            "(e.g., 'Development Status :: 3 - Alpha'), but found none.",
            file=sys.stderr,
        )
        success = False
    else:
        expected_status_in_init = status_map[status_classifier]

        if expected_status_in_init != aas_core3_1.__status__:
            print(
                f"Expected status {expected_status_in_init} "
                f"according to pyproject.toml in aas_core3_1/__init__.py, "
                f"but found: {aas_core3_1.__status__}"
            )
            success = False

    if not success:
        return -1

    return 0


if __name__ == "__main__":
    sys.exit(main())



def data_specification_iec_61360s_have_value(
    embedded_data_specifications: Iterable[aas_types.EmbeddedDataSpecification],
) -> bool:
    """
    Check that :py:attr:`.types.DataSpecificationIEC61360.value`
    is defined for all data specifications whose content is given as IEC 61360.
    """
    for embedded_data_specification in embedded_data_specifications:
        if isinstance(
            embedded_data_specification.data_specification_content,
            aas_types.DataSpecificationIEC61360,
        ):
            iec61360 = embedded_data_specification.data_specification_content
            if iec61360.value is None:
                return False

    return True



def id_shorts_of_variables_are_unique(
    input_variables: Optional[List[aas_types.OperationVariable]],
    output_variables: Optional[List[aas_types.OperationVariable]],
    inoutput_variables: Optional[List[aas_types.OperationVariable]],
) -> bool:
    """
    Check that the :py:attr:`.types.Referable.id_short`'s among all the
    :paramref:`input_variables`, :paramref:`output_variables`
    and :paramref:`inoutput_variables` are unique.
    """
    id_short_set = set()
    if input_variables is not None:
        for variable in input_variables:
            if variable.value.id_short is not None:
                if variable.value.id_short in id_short_set:
                    return False

                id_short_set.add(variable.value.id_short)
    if output_variables is not None:
        for variable in output_variables:
            if variable.value.id_short is not None:
                if variable.value.id_short in id_short_set:
                    return False

                id_short_set.add(variable.value.id_short)
    if inoutput_variables is not None:
        for variable in inoutput_variables:
            if variable.value.id_short is not None:
                if variable.value.id_short in id_short_set:
                    return False

                id_short_set.add(variable.value.id_short)
    return True



# fmt: off
_AAS_SUBMODEL_ELEMENTS_TO_TYPE: Mapping[
    aas_types.AASSubmodelElements,
    type
] = {
    aas_types.AASSubmodelElements.ANNOTATED_RELATIONSHIP_ELEMENT:
        aas_types.AnnotatedRelationshipElement,

    aas_types.AASSubmodelElements.BASIC_EVENT_ELEMENT:
        aas_types.BasicEventElement,

    aas_types.AASSubmodelElements.BLOB:
        aas_types.Blob,

    aas_types.AASSubmodelElements.CAPABILITY:
        aas_types.Capability,

    aas_types.AASSubmodelElements.DATA_ELEMENT:
        aas_types.DataElement,

    aas_types.AASSubmodelElements.ENTITY:
        aas_types.Entity,

    aas_types.AASSubmodelElements.EVENT_ELEMENT:
        aas_types.EventElement,

    aas_types.AASSubmodelElements.FILE:
        aas_types.File,

    aas_types.AASSubmodelElements.MULTI_LANGUAGE_PROPERTY:
        aas_types.MultiLanguageProperty,

    aas_types.AASSubmodelElements.OPERATION:
        aas_types.Operation,

    aas_types.AASSubmodelElements.PROPERTY:
        aas_types.Property,

    aas_types.AASSubmodelElements.RANGE:
        aas_types.Range,

    aas_types.AASSubmodelElements.REFERENCE_ELEMENT:
        aas_types.ReferenceElement,

    aas_types.AASSubmodelElements.RELATIONSHIP_ELEMENT:
        aas_types.RelationshipElement,

    aas_types.AASSubmodelElements.SUBMODEL_ELEMENT:
        aas_types.SubmodelElement,

    aas_types.AASSubmodelElements.SUBMODEL_ELEMENT_LIST:
        aas_types.SubmodelElementList,

    aas_types.AASSubmodelElements.SUBMODEL_ELEMENT_COLLECTION:
        aas_types.SubmodelElementCollection,
}
# fmt: on


def _assert_all_types_covered_in_aas_submodel_elements_to_type() -> None:
    """
    Assert that we did not miss a type in :py:attr:`_AAS_SUBMODEL_ELEMENTS_TO_TYPE`.
    """
    missing_literals = [
        literal
        for literal in aas_types.AASSubmodelElements
        if literal not in _AAS_SUBMODEL_ELEMENTS_TO_TYPE
    ]

    assert len(missing_literals) == 0, (
        f"Some literals were missed in "
        f"_AAS_SUBMODEL_ELEMENTS_TO_TYPE: {missing_literals!r}"
    )


_assert_all_types_covered_in_aas_submodel_elements_to_type()


def submodel_element_is_of_type(
    element: aas_types.SubmodelElement, expected_type: aas_types.AASSubmodelElements
) -> bool:
    """
    Check that :paramref:`element` is an instance of class corresponding
    to :paramref:`expected_type`.
    """
    # noinspection PyTypeHints
    return isinstance(element, _AAS_SUBMODEL_ELEMENTS_TO_TYPE[expected_type])



def data_specification_iec_61360s_have_definition_at_least_in_english(
    embedded_data_specifications: Iterable[aas_types.EmbeddedDataSpecification],
) -> bool:
    """
    Check that :py:attr:`.types.DataSpecificationIEC61360.definition`
    is defined for all data specifications whose content is given as IEC 61360
    at least in English.
    """
    for embedded_data_specification in embedded_data_specifications:
        if isinstance(
            embedded_data_specification.data_specification_content,
            aas_types.DataSpecificationIEC61360,
        ):
            iec61360 = embedded_data_specification.data_specification_content
            if iec61360.definition is None:
                return False

            no_definition_in_english = True
            for lang_string in iec61360.definition:
                if is_bcp_47_for_english(lang_string.language):
                    no_definition_in_english = False
                    break

            if no_definition_in_english:
                return False

    return True



def reference_key_values_equal(
    that: aas_types.Reference, other: aas_types.Reference
) -> bool:
    """
    Check that the two references, :paramref:`that` and :paramref:`other`,
    are equal by comparing their :py:attr:`.types.Reference.keys`
    by :py:attr:`.types.Key.value`'s.
    """
    if len(that.keys) != len(other.keys):
        return False

    for that_key, other_key in zip(that.keys, other.keys):
        if that_key.value != other_key.value:
            return False

    return True
